/* AsmReader.cpp */

#include "AsmReader.hpp"
#include <string>
#include <regex>
#include <functional>

bool onmatch(std::string line, std::string rx, std::function<void(std::smatch)> handler) {
	std::smatch matches;
	std::regex r(rx, std::regex_constants::icase);

	if (std::regex_match(line, matches, r)) {
		handler(matches);
		return true;

	} else {
		return false;
	}
}

int32_t parsenum(std::string input) {
	if (input.length() >= 3 && input[0] == '0' && input[1] == 'x') {
		return std::strtol(input.c_str() + 2, nullptr, 16);

	} else {
		return std::atoi(input.c_str());

	}
}

Register lookupRegister(std::string input) {
	if (input == "eax") {
		return REG_A;
	} else if (input == "ecx") {
		return REG_C;
	} else if (input == "edx") {
		return REG_D;
	} else if (input == "ebx") {
		return REG_B;
	} else if (input == "esp") {
		return REG_SP;
	} else if (input == "ebp") {
		return REG_BP;
	} else if (input == "esi") {
		return REG_SI;
	} else if (input == "edi") {
		return REG_DI;
	} else {
		return REG__INVALID;
	}
}

bool AsmReader::process(std::istream* input, std::ostream* output, AsmBuilder* ab) {
	
	unsigned int line_count = 0;

	std::string line;
	while (!input->eof()) {
		std::getline(*input, line);
		line_count++;

		if (!line.length()) {
			continue;
		}

		// Strip leading whitespace
		auto nonspacepos = line.find_first_not_of(" \t\r\v");
		if (nonspacepos != line.npos && nonspacepos > 0) {
			line = line.substr(nonspacepos);
		}

		// Strip trailing comments
		auto commentpos = line.find_first_of(';');
		if (commentpos != line.npos) {
			line = line.substr(0, commentpos);
		}

		// Strip trailing whitespace
		auto rspacepos = line.find_last_not_of(" \t\r\v");
		if (rspacepos != line.npos) {
			line.erase(rspacepos + 1);
		} else {
			continue; // entirely whitespace
		}

		//

#define WSP_BASE "[\\s\\t]"
#define WSP WSP_BASE "+"
#define OSP WSP_BASE "*"
#define ARGSEP WSP_BASE "*," WSP_BASE "*"
#define NUM "(0x[0-9a-f]+|-?\\d+)"
#define LABELSTR "([a-zA-Z0-9_-]+)"

		if (
			// single-line no-ops

			!(line == "bits 32") &&
			!(line == "BITS 32") &&

			// regex matches

			!onmatch(line, "^.db" WSP NUM "$", [&](std::smatch s) {
				int32_t arg = parsenum(s[1].str());
				uint8_t arg8 = arg & 0xFF;
				ab->raw8(arg8);
			}) &&

			!onmatch(line, "^.dw" WSP NUM "$", [&](std::smatch s) {
				int32_t arg = parsenum(s[1].str());
				uint16_t arg16 = arg & 0xFFFF;
				ab->raw16(arg16);
			}) &&

			!onmatch(line, "^.dd" WSP NUM "$", [&](std::smatch s) {
				int32_t arg = parsenum(s[1].str());
				uint32_t arg32 = arg;
				ab->raw32(arg32);
			}) &&

			!onmatch(line, "^push" WSP "(e..)$", [&](std::smatch s) {
				ab->pushReg(lookupRegister(s[1]));
			}) &&

			!onmatch(line, "^pop" WSP "(e..)$", [&](std::smatch s) {
				ab->popReg(lookupRegister(s[1]));
			}) &&

			!onmatch(line, "^push" WSP "dword" OSP "\\[" OSP "(e..)" OSP "\\]$", [&](std::smatch s) {
				ab->pushRegOffset(lookupRegister(s[1]), 0);
			}) &&

			!onmatch(line, "^push" WSP "dword" OSP "\\[" OSP "(e..)" OSP "([+-])" OSP NUM OSP "\\]$", [&](std::smatch s) {
				int32_t arg = parsenum(s[3].str());
				if (s[2].str()[0] == '-') {
					arg = -arg;
				}
				ab->pushRegOffset(lookupRegister(s[1]), arg);
			}) &&

			!onmatch(line, "^mov" WSP "(e..)" ARGSEP "(e..)$", [&](std::smatch s) {
				ab->movRegReg(lookupRegister(s[1]), lookupRegister(s[2]));
			}) &&

			!onmatch(line, "^mov" WSP "(e..)" ARGSEP NUM "$", [&](std::smatch s) {
				int32_t arg = parsenum(s[2].str());
				ab->movRegImm32(lookupRegister(s[1]), arg);
			}) &&

			!onmatch(line, "^mov" WSP "(e..)" ARGSEP "dword" OSP "\\[" OSP "(e..)" OSP "\\]$", [&](std::smatch s) {
				ab->movRegRegOffset(lookupRegister(s[1]), lookupRegister(s[2]), 0);
			}) &&

			!onmatch(line, "^mov" WSP "(e..)" ARGSEP "dword" OSP "\\[" OSP "(e..)" OSP "([+-])" OSP NUM OSP "\\]$", [&](std::smatch s) {
				int32_t arg = parsenum(s[4].str());
				if (s[3].str()[0] == '-') {
					arg = -arg;
				}
				ab->movRegRegOffset(lookupRegister(s[1]), lookupRegister(s[2]), arg);
			}) &&

			!onmatch(line, "^mov" WSP "dword" OSP "\\[" OSP "(e..)" OSP "\\]" ARGSEP "(e..)$", [&](std::smatch s) {
				ab->movRegOffsetReg(lookupRegister(s[1]), 0, lookupRegister(s[2]));
			}) &&

			!onmatch(line, "^mov" WSP "dword" OSP "\\[" OSP "(e..)" OSP "([+-])" OSP NUM OSP "\\]" ARGSEP "(e..)$", [&](std::smatch s) {
				int32_t arg = parsenum(s[3].str());
				if (s[2].str()[0] == '-') {
					arg = -arg;
				}
				ab->movRegOffsetReg(lookupRegister(s[1]), arg, lookupRegister(s[4]));
			}) &&

			!onmatch(line, "^add" WSP "(e..)" ARGSEP "(e..)$", [&](std::smatch s) {
				ab->addRegReg(lookupRegister(s[1]), lookupRegister(s[2]));
			}) &&

			!onmatch(line, "^add" WSP "(e..)" ARGSEP "(\\d+)$", [&](std::smatch s) {
				int32_t arg = parsenum(s[2].str());
				ab->addRegImm32(lookupRegister(s[1]), arg);
			}) &&

			!onmatch(line, "^xor" WSP "(e..)" ARGSEP "(e..)$", [&](std::smatch s) {
				Register acc = lookupRegister(s[1]), src = lookupRegister(s[2]);
				ab->clearRegister(acc == src ? acc : REG__INVALID); // only clear is supported
			}) &&

			!onmatch(line, "^" LABELSTR ":$", [&](std::smatch s) {
				ab->label(s[1]);
			}) &&

			!onmatch(line, "^global" WSP LABELSTR "$", [&](std::smatch s) {
				ab->exportGlobalHere(s[1]);
			}) && 

			// Jumps

			!onmatch(line, "^jmp" WSP LABELSTR "$", [&](std::smatch s) {
				ab->jumpUnconditional(s[1]);
			}) && 

			!onmatch(line, "^(?:jle|jng)" WSP LABELSTR "$", [&](std::smatch s) {
				ab->jumpIfLesserOrEqual(s[1]);
			}) &&

			!onmatch(line, "^(?:jl|jnge)" WSP LABELSTR "$", [&](std::smatch s) {
				ab->jumpIfLesser(s[1]);
			}) &&

			!onmatch(line, "^j[ez]" WSP LABELSTR "$", [&](std::smatch s) {
				ab->jumpIfEqual(s[1]);
			}) &&

			!onmatch(line, "^jn[ez]" WSP LABELSTR "$", [&](std::smatch s) {
				ab->jumpIfNotEqual(s[1]);
			}) &&

			!onmatch(line, "^(?:jg|jnle)" WSP LABELSTR "$", [&](std::smatch s) {
				ab->jumpIfGreater(s[1]);
			}) &&

			!onmatch(line, "^(?:jge|jnl)" WSP LABELSTR "$", [&](std::smatch s) {
				ab->jumpIfGreaterOrEqual(s[1]);
			}) &&

			// Noop

			!onmatch(line, "^noo?p$", [&](std::smatch s) {
				ab->noop();
			}) &&

			// Calls

			!onmatch(line, "^call" WSP "(e..)$", [&](std::smatch s) {
				ab->callFromRegister(lookupRegister(s[1]));
			}) &&

			!onmatch(line, "^call" WSP LABELSTR "$", [&](std::smatch s) {
				ab->callByLabel(s[1]);
			}) &&

			!onmatch(line, "^leave$", [&](std::smatch s) {
				ab->leave();
			}) &&

			!onmatch(line, "^retn$", [&](std::smatch s) {
				ab->returnFunc();
			}) &&

			!onmatch(line, "^retn" WSP NUM "$", [&](std::smatch s) {
				int32_t arg = parsenum(s[1].str());
				uint16_t arg16 = arg;
				ab->returnFunc(arg16);
			}) &&

			// .

			!onmatch(line, "^section" WSP "(.+)$", [&](std::smatch s) {
				ab->changeSection(s[1]);
			}) &&

			true
		) {
			// None of our patterns matched for this line
			errors << "Unknown opcode at line " << line_count << ": '" << line << "'\n";
			return false;
		}

	}

	// EOF reached

	std::string tmp;
	if (ab->renderObject(tmp)) {
		*output << tmp;
		return true;

	} else {
		errors << "Failed to render output\n";
		return false;

	}
}
